package edu.uci.ics.pattis.ics23.collections;


import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;



//put           : O(N)
//putAll        : O(MN) (M values in Iterable/Collection) 
//remove        : O(N)
//clear         : O(1)   with garbage collection issue
//get           : O(N)
//containsKey   : O(N)
//containsValue : O(N)
//entries       : O(1) 
//keys          : O(1)
//values        : O(1)
//  hasNext     : O(1) For entries, keys, and values
//  next        : O(1)
//  remove      : O(1)
//isEmpty       : O(1)
//size          : O(1)
//toArray       : O(N)
//newEmpty      : O(1)
//shallowCopy   : O(N)
//toString      : O(N)
//equals        : O(N^2)
//hashCode      : O(N)



public class ArrayMap<K,V> extends AbstractMap<K,V> implements Map<K,V> {

	
  public ArrayMap()
  {this(1);}

	
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArrayMap(int initialCapacity) {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException
        ("ArrayMap("+initialCapacity+"): illegal initialCapacity");

    map = (Entry<K,V>[])new Map.Entry[initialCapacity];
 }


  public ArrayMap (Iterable<Map.Entry<K,V>> keysValues) {
    this(1);
    putAll(keysValues);
  }


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArrayMap (Map.Entry<K,V>[] keysValues) {
    map = (Map.Entry<K,V>[])new Map.Entry[Math.max(1,keysValues.length)]; 
    for (Map.Entry<K,V> e : keysValues)
      put(e.getKey(),e.getValue());
  } 

	
	 
  //This method is required for a concrete class implementing Map
  public V put (K key, V value){
    int i = indexOf(key);
    if (i != -1)
      return changeAt(i,value);

    ensureCapacity(objectCount+1);
    map[objectCount++] = new AbstractMap.SimpleEntry<K,V>(key,value);
    modCount++;
    return null;
  }
	
	
	
  //This method improves the performance of the method inherited
  //  from AbstractMap
  public V remove (K k){
    int i = indexOf(k);
    if (i != -1)
      return removeAt(i);
    return null;
  }
	
	
	
  //This method improves the performance of the method inherited
  //  from AbstractMap
  public void clear() {
    //Save time by not putting null everywhere in array
    //for (int i=0; i<objectCount; i++)
    //  map[i] = null;
    objectCount = 0;
    modCount++;
  }

	
  //This method improves the performance of the method inherited
  //  from AbstractMap
  public V get (Object key) {
    int i = indexOf(key);
    if (i != -1)
      return map[i].getValue();
    return null;
  }
	
	
	
  //This method improves the performance of the method inherited
  //  from AbstractMap
  public boolean containsKey (Object key)
  {return indexOf(key) != -1;}
	
	
	
  //This method is required for a concrete class implementing Map
  public  Iterable<Map.Entry<K,V>> entries () {
    return new Iterable<Map.Entry<K,V>>() {
  	public Iterator<Map.Entry<K,V>> iterator() {
  	  return new MapEntryIterator();
        }
    };
  }
  
  
  
  //This method improves the performance of the method inherited
  //  from AbstractMap
  public int size ()
  {return objectCount;}
	
	
	
  //This method is required for a concrete class implementing List
  public Map<K,V> newEmpty (){
    return new ArrayMap<K,V>();
  }

  
  
  //This method is required for a concrete class implementing Map
  public Map<K,V>  shallowCopy() {
    //Must construct object and set all instance variables appropriately
    ArrayMap<K,V> answer = new ArrayMap<K,V>(map.length);
    answer.objectCount = objectCount;
    for (int i=0; i<objectCount; i++)
      answer.map[i] = map[i];
    return answer;
  }
 
	
	
  //Could use the inherited iterator version, adding extra data structure information
  //public String toString()
  //{return "ArrayMap["+objectCount+"/"+map.length+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArrayMap["+objectCount+"/"+map.length+":");
    for (int i=0; i<objectCount; i++)
       answer.append( "["+i+"]=" +
                      (map[i] == this ? "(this Map)" : String.valueOf(map[i])) +
                      (i!=objectCount-1 ? "," : "")
                     );
    answer.append("]");
     
    return answer.toString();
  }
  
  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = map.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      Map.Entry<K,V>[] old = map;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
        newCapacity = minCapacity;
       map = (Map.Entry<K,V>[])new Map.Entry[newCapacity];
       for (int i=0; i<objectCount; i++)
         map[i] = old[i];
    }
  } 
   


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = map.length;
    if (objectCount < oldCapacity) {
      modCount++;
      Map.Entry<K,V>[] old = map;
      map = (Map.Entry<K,V>[])new Map.Entry[objectCount];
      for (int i=0; i<objectCount; i++)
 	map[i] = old[i];
     }
  }
   
   
   
  private int indexOf(Object o) {
    if (o == null) {
      for (int i = 0; i<objectCount; i++)
        if (map[i].getKey() == null)
          return i;
    } else
      for (int i = 0; i<objectCount; i++)
        if (o.equals(map[i].getKey()))
          return i;
    
    return -1;
  }
  
 
  
  private V removeAt(int i) {
    V value          = map[i].getValue();
    map[i]           = map[--objectCount]; 	
    map[objectCount] = null;
    modCount++;
    return value;
  }

  
  
  private V changeAt(int i, V value) {
    V oldValue = map[i].getValue();
    map[i].setValue(value); 	
    return oldValue;
  }
  
 
  
  //Concrete data structure for implementing a Map via an array
  //The order of the values in the array IS NOT important
  private Map.Entry<K,V>[] map;
  private int objectCount;
 
  

  
  private class MapEntryIterator implements Iterator<Map.Entry<K,V>> {

    public boolean hasNext()
    {return nextIteratorIndex < objectCount;}

    
    public Map.Entry<K,V> next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayMap.MapEntryIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArrayMap.MapEntryIterator.next");

      Map.Entry<K,V> answer = map[nextIteratorIndex++];
      removedAlready = false;
      return answer;
    }

    
    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayMap.MapEntryIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArrayMap.MapEntryIterator.remove");

      removeAt(nextIteratorIndex-1);
      nextIteratorIndex--;
      expectedModCount = modCount;  //removeAt does: objectCount-- modCount++
      removedAlready   = true;
    }

    
    private int     nextIteratorIndex  = 0;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }

}
