package edu.uci.ics.pattis.ics23.collections;


import java.util.Iterator;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;



//add         : O(1) amortized
//addAll      : O(M) (M values in Iterable), amortized
//remove      : O(1)
//clear       : O(1) with garbage collection issue
//peek        : O(1)
//isEmpty     : O(1)
//size        : O(1)
//iterator    : O(1)
//  hasNext   : O(1)
//  next      : O(1)
//  remove    : O(N)
//toArray     : O(N)
//newEmpty    : O()
//shallowCopy : O(N)
//toString    : O(N)
//equals      : O(N)
//hashCode    : O(N)



public class ArrayQueue<E> extends AbstractQueue<E> implements Queue<E> {
    

  public ArrayQueue ()
  {this(1);}
  

  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArrayQueue (int initialCapacity) {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException
        ("ArrayQueue("+initialCapacity+"): illegal initialCapacity");
 
    q = (E[])new Object[initialCapacity];
  }

  
  public ArrayQueue (Iterable<? extends E> es) {
    this(1);
    addAll(es);
  }
  

  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArrayQueue (E[] es) {
    q = (E[])new Object[Math.max(1,es.length)]; 
    for (E e : es)
      add(e);
  } 
  
  
  

  //This method is required for a concrete class implementing Queue
  public boolean add (E e) {
    ensureCapacity(objectCount+1);
    q[last= (last+1)%q.length] = e;
    objectCount++;
    modCount++;
    return true;
  }
  

  
  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public E remove () throws NoSuchElementException {
    if (isEmpty())
       throw new NoSuchElementException("ArrayQueue.remove: empty queue");

    E answer = q[first];
    q[first] = null;
    first    = (first+1)%q.length;
    objectCount--;
    modCount++;
    return answer;
  }

  
  
  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public void clear () {
    //Save time by not putting null everywhere in array
    //for (int i=0; i<objectCount; i++)
    //  q[(first+i)%q.length] = null;
    objectCount = 0;
    first = 0;
    last = -1;
    modCount++;
  }



  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public E peek () throws NoSuchElementException {
    if (isEmpty())
      throw new NoSuchElementException("ArrayQueue.peek: empty queue");

    return q[first];
  }



  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
   public int size ()
   {return objectCount;}
       
    

  //This method is required for a concrete class implementing Queue
  public Iterator<E> iterator()
  {return new ArrayQueueIterator();}
   
    
    

  //This method is required for a concrete class implementing List
  public Queue<E> newEmpty (){
    return new ArrayQueue<E>();
  }

   
   
  //This method is required for a concrete class implementing Queue
  public Queue<E>  shallowCopy() {
    //Must construct object and set all instance variables appropriately
    ArrayQueue<E> answer = new ArrayQueue<E>(q.length);
    answer.objectCount = objectCount;
    answer.first       = first;
    answer.last        = last;
    for (int i=0; i<objectCount; i++)
      answer.q[(first+i)%q.length] = q[i];
    return answer;
  }
   
   
   
  //Could use the inherited iterator version, adding extra data structure information
  //public String toString()
  //{return "ArrayQueue["+objectCount+"/"+q.length+"/"+first+"/"+last+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArrayQueue["+objectCount+"/"+q.length+"/"+first+"/"+last+":");
    for (int i=0; i<objectCount; i++)
      answer.append( "["+(i+first)%q.length+"]=" +
                     (q[(first+i)%q.length] == this ? "(this Collection)" : String.valueOf(q[(first+i)%q.length])) + 
                     (i!=objectCount-1 ? "," : "")
                    );
    answer.append("]");
    
    return answer.toString();
  }



  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = q.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      E[] old = q;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
        newCapacity = minCapacity;
      q = (E[])new Object[newCapacity];
      for (int i=0; i<objectCount; i++)
 	q[i] = old[(first+i)%old.length];
      first = 0;
      last  = objectCount-1;
    }
  } 
   


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = q.length;
    if (objectCount < oldCapacity) {
      modCount++;
      E[] old = q;
      q = (E[])new Object[objectCount];
      for (int i=0; i<objectCount; i++)
        q[i] = old[(first+i)%old.length];
      first = 0;
      last  = objectCount-1;
    }
  }
   
   
   
   
  //Concrete data structure for implementing a Queue via an array
  //The order of the values in the array IS important in this "circular" array
  private int first = 0;
  private int last = -1;
  private E[] q;
  private int objectCount;
  

  

  private class ArrayQueueIterator implements Iterator<E> {

    public boolean hasNext()
    {return leftToIterate != 0;}


    public E next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayQueueIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArrayQueueIterator.next");

      E answer = q[nextIteratorIndex];
      nextIteratorIndex = (nextIteratorIndex+1)%q.length;
      leftToIterate--;
      removedAlready = false;
      return answer;
    }


    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayQueueIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArrayQueueIterator.remove");

      int to = (nextIteratorIndex != 0 ? nextIteratorIndex-1 : q.length-1);
      if (to == first)
        first++;
      else {
        int from = nextIteratorIndex;
        for (int i=0; i<leftToIterate; i++) {
          q[to] = q[from];
          to    =(to+1  )%q.length;
          from  =(from+1)%q.length;
        }
        last = (last != 0 ? last-1 : q.length-1);
        nextIteratorIndex = (nextIteratorIndex != 0 ? nextIteratorIndex-1 : q.length-1);
      }

      q[to] = null;
      objectCount--;
      expectedModCount = ++modCount;
      removedAlready   = true;
    }

    
    private int     leftToIterate      = objectCount;
    private int     nextIteratorIndex  = first;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }
}
