package edu.uci.ics.pattis.ics23.collections;


import java.util.Iterator;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;



//add         : O(1) amortized
//addAll      : O(M) (M values in Iterable), amortized
//remove      : O(1)
//clear       : O(1) with garbage collection issue
//peek        : O(1)
//isEmpty     : O(1)
//size        : O(1)
//iterator    : O(1)
//  hasNext   : O(1)
//  next      : O(1)
//  remove    : O(N)
//toArray     : O(N)
//newEmpty    : O(1)
//shallowCopy : O(N)
//toString    : O(N)
//equals      : O(N)
//hashCode    : O(N)



public class ArrayStack<E> extends AbstractStack<E> implements Stack<E> {
    


  public ArrayStack ()
  {this(1);}

 
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it contains no objects
  @SuppressWarnings("unchecked")
  public ArrayStack (int initialCapacity) {
    if (initialCapacity <= 0)
      throw new IllegalArgumentException
        ("ArrayStack("+initialCapacity+"): illegal initialCapacity");
 
    stack = (E[]) new Object[initialCapacity];
 }
  

  public ArrayStack (Iterable<? extends E> es) {
    this(1);
    addAll(es);
  }

  
  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public ArrayStack (E[] es) {
    stack = (E[])new Object[Math.max(1,es.length)]; 
    for (E e : es)
      add(e);
  } 
  
	
	
  //This method is required for a concrete class implementing Stack
  public boolean add (E e) {
    ensureCapacity(objectCount+1);
    stack[objectCount++] = e;
    modCount++;
    return true;
  }
  

  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public E remove () throws NoSuchElementException {
    if (isEmpty())
      throw new NoSuchElementException("ArrayStack.remove: empty stack");

    E answer           = stack[--objectCount];
    stack[objectCount] = null;
    modCount++;
    return answer;
 }
  
  
  
  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public void clear () {
    //Save time by not putting null everywhere in array
    //for (int i=0; i<objectCount; i++)
    //  stack[i] = null;
    objectCount = 0;
    modCount++;
  }



  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public E peek () throws NoSuchElementException {
    if (isEmpty())
      throw new NoSuchElementException("ArrayStack.peek: empty stack");

    return stack[objectCount-1];
  }


  //This method improves the performance of the method inherited
  //  from AbstractOrderedCollection
  public int size ()
  {return objectCount;}
 


  //This method is required for a concrete class implementing Stack
  public Iterator<E> iterator ()
  {return new ArrayStackIterator();}
  
   

  //This method is required for a concrete class implementing List
  public Stack<E> newEmpty (){
    return new ArrayStack<E>();
  }

  
  
  //This method is required for a concrete class implementing Stack
  public Stack<E>  shallowCopy() {
    //Must construct object and set all instance variables appropriately
    ArrayStack<E> answer = new ArrayStack<E>(stack.length);
    answer.objectCount = objectCount;
    for (int i=0; i<objectCount; i++)
      answer.stack[i] = stack[i];
    return answer;
  }
  
  
  
  //Could use the inherited iterator version, adding extra data structure information
  //public String toString()
  //{return "ArrayStack["+objectCount+"/"+stack.length+":"+super.toString()+"]";}
 
  //Directly use information about the data structure
  public String toString() {
    StringBuffer answer = new StringBuffer("ArrayStack["+objectCount+"/"+stack.length+":");
    for (int i=0; i<objectCount; i++)
      answer.append( "["+i+"]=" +
                     (stack[i] == this ? "(this Collection)" : String.valueOf(stack[i])) + 
                     (i!=objectCount-1 ? "," : "")
                    );
    answer.append("]");
    
    return answer.toString();
  }



  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void ensureCapacity (int minCapacity) {
    int oldCapacity = stack.length;
    if (minCapacity > oldCapacity) {
      modCount++;
      Object old[] = stack;
      int newCapacity = 2*oldCapacity;
      if (newCapacity < minCapacity)
	 newCapacity = minCapacity;
      stack = (E[]) new Object[newCapacity];
      System.arraycopy(old,0,stack,0,objectCount);
    }
  } 
  


  //It is OK that the new constructed Object[] is cast to E[]
  //  because it is filled with E objects
  @SuppressWarnings("unchecked")
  public void trimToSize () {
    int oldCapacity = stack.length;
    if (objectCount < oldCapacity) {
      modCount++;
      Object old[] = stack;
      stack = (E[]) new Object[objectCount];
      System.arraycopy(old,0,stack,0,objectCount);
    }
  }
  
  
  

  //Concrete data structure for implementing a Stack via an array
  //The order of the values in the array IS important (bottom to top)
  private E[] stack;
  private int objectCount = 0;

 
  
  
  private class ArrayStackIterator implements Iterator<E> {

    public boolean hasNext()
    {return nextIteratorIndex != -1;}


    public E next() throws ConcurrentModificationException, NoSuchElementException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayStackIterator.next");
      if (!hasNext())
        throw new NoSuchElementException("ArrayStackIterator.next");

      removedAlready = false;
      return stack[nextIteratorIndex--];
    }


    public void remove() throws ConcurrentModificationException, UnsupportedOperationException, IllegalStateException {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException("ArrayStackIterator.remove");
      if (removedAlready)
        throw new IllegalStateException("ArrayStackIterator.remove");

      for (int i=nextIteratorIndex+1; i<objectCount-1; i++)
        stack[i] = stack[i+1];
      stack[--objectCount] = null;
      expectedModCount     = ++modCount;
      removedAlready       = true;
    }


    private int     nextIteratorIndex  = objectCount-1;
    private long    expectedModCount   = modCount;
    private boolean removedAlready     = true;
  }

}
